package com.persagy.dmp.admin.service;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.JsonNodeFactory;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.persagy.common.json.JacksonMapper;
import com.persagy.common.web.MapResponse;
import com.persagy.dmp.admin.entity.TransObjectInfo;
import com.persagy.dmp.admin.repository.TransObjectInfoRepository;
import com.persagy.dmp.admin.service.trans.RwdTransUtils;
import com.persagy.dmp.admin.service.trans.TransObjectContext;
import com.persagy.dmp.common.http.HttpUtils;
import com.persagy.dmp.rwd.model.DmpMessage;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;

import javax.transaction.Transactional;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * @author: yaoll
 * @date: 2020-12-30
 * @verison: 1.0
 */
@Slf4j
@Service
public class DmpTransObjectService {

    private static final String KEY_TRANS = "admin_trans_";

    @Autowired
    private RwdTransUtils transUtils;

    @Autowired
    private HttpUtils httpUtils;

    @Value("${persagy.dmp.server}")
    private String dmpServer;

    @Autowired
    private StringRedisTemplate redis;

    @Autowired
    private TransObjectInfoRepository objectInfoRepository;

    // 更新数据
    @Transactional
    public MapResponse againTrans(String groupCode, String projectId, List<TransObjectInfo> transObjectInfoList) {
        MapResponse resp = new MapResponse();
        log.info("object again trans start size: {}", transObjectInfoList.size());

        Map<String, List<TransObjectInfo>> objectMap = new HashMap<>();
        for (TransObjectInfo transObjectInfo : transObjectInfoList) {
            String objectId = transObjectInfo.getObjectId();
            if (objectId != null) {
                objectMap.computeIfAbsent(objectId, k -> new ArrayList<>());
                objectMap.get(objectId).add(transObjectInfo);
            }
        }
        if (objectMap.size() == 0) {
            return resp;
        }

        String key = KEY_TRANS + projectId;
        Long flag = redis.opsForValue().increment(key, 1L);
        if (flag > 1) {
            resp.setFail("正在处理中...");
            return resp;
        }
        redis.expire(key, 24, TimeUnit.HOURS);

        for (List<TransObjectInfo> value : objectMap.values()) {
            objectInfoRepository.deleteAll(value);
        }

        TransObjectContext ctx = new TransObjectContext();
        ctx.setAppId("datautils");
        ctx.setUserId("system");
        ctx.setGroupCode(groupCode);
        ctx.setProjectId(projectId);
        List<TransObjectInfo> logs = new ArrayList<>();
        objectMap.values().forEach(logs::addAll);
        //启动
        start(ctx);
        ctx.setLogs(logs);
        initClass(ctx);
        // 更新数据到中台
        transUtils.updateObject(ctx, objectMap);
        return resp;
    }

    // 开始迁移
    @Transactional
    public MapResponse trans(Map<String, Object> param) {
        MapResponse resp = new MapResponse();
        String groupCode = MapUtils.getString(param, "groupCode");
        if (groupCode == null) {
            resp.setFail("groupCode is required");
            return resp;
        }
        String projectId = MapUtils.getString(param, "projectId");
        if (projectId == null) {
            resp.setFail("projectId is required");
            return resp;
        }
        String server = MapUtils.getString(param, "server");
        if (server == null) {
            resp.setFail("server is required");
            return resp;
        }
        String secret = MapUtils.getString(param, "secret");
        if (secret == null) {
            resp.setFail("secret is required");
            return resp;
        }
        log.info("object trans start: {}", JacksonMapper.toSimpleJson(param));

        String key = KEY_TRANS + projectId;
        Long flag = redis.opsForValue().increment(key, 1L);
        if (flag > 1) {
            resp.setFail("正在处理中...");
            return resp;
        }
        redis.expire(key, 24, TimeUnit.HOURS);

        objectInfoRepository.deleteByProjectId(projectId);

        DmpMessage message = new DmpMessage("transObject", projectId);
        message.setGroupCode(groupCode);
        message.setProjectId(projectId);
        message.setStr1(server);
        message.setStr2(secret);
        resp.add(message);
        return resp;
    }

    // 开始标记
    public void start(TransObjectContext ctx) {
        transUtils.writeInfo(ctx, "info", "start");
    }

    // 初始化类型
    public void initClass(TransObjectContext ctx) {
        transUtils.initClass(ctx);
    }

    // 迁移项目
    public void transProject(TransObjectContext ctx) {
        transByType(ctx, "project", "Pj");
    }

    // 迁移建筑
    public void transBuilding(TransObjectContext ctx) {
        transByType(ctx, "building", "Bd");
    }

    // 迁移楼层
    public void transFloor(TransObjectContext ctx) {
        transByType(ctx, "floor", "Fl");
    }

    // 迁移竖井
    public void transShaft(TransObjectContext ctx) {
        transByType(ctx, "shaft", "Sh");
    }

    private void transByType(TransObjectContext ctx, String objType, String type) {
        String classCode = ctx.getClassCodeMapping().get(type);
        if (classCode == null) {
            transUtils.writeInfo(ctx, "error", "类型[" + type + "]未定义");
            return;
        }
        List<ObjectNode> list = transUtils.queryObject(ctx, type);
        if (list == null) {
            return;
        }
        // 加载信息点
        transUtils.initFuncidCache(ctx, classCode);
        List<ObjectNode> objects = transUtils.toDmpObject(ctx, list, objType, classCode);
        transUtils.save(ctx, objects, type);
    }

    // 迁移空间
    public void transSpace(TransObjectContext ctx) {
        String objType = "space";
        List<ObjectNode> list = transUtils.queryObject(ctx, "Sp");
        if (list == null) {
            return;
        }
        Set<String> businessTypes = new HashSet<>();
        for (ObjectNode obj : list) {
            businessTypes.add(obj.get("businessType").asText());
        }

        for (String classCode : businessTypes) {
            // 加载信息点
            transUtils.initFuncidCache(ctx, classCode);

            List<ObjectNode> data = new LinkedList<>();
            for (ObjectNode obj : list) {
                String businessType = JacksonMapper.getString(obj, "businessType");
                if (classCode.equalsIgnoreCase(businessType)) {
                    obj.remove("businessType");
                    data.add(obj);
                }
            }
            List<ObjectNode> objects = transUtils.toDmpObject(ctx, data, objType, classCode);
            transUtils.save(ctx, objects, classCode);
        }
    }

    // 迁移系统设备部件
    public void transEquipment(TransObjectContext ctx) {
        String url = ctx.getServer() + "/data-platform-3/dict/query/equipment_all?project_id=" + ctx.getProjectId();
        HttpUtils.HttpResult httpResult = httpUtils.get(url, 300 * 1000);
        if (!httpResult.success()) {
            transUtils.writeInfo(ctx, "error", "查询系统设备部件类型失败");
            return;
        }
        String content = httpResult.getContent();
        ObjectNode objectNode = JacksonMapper.toObject(content, ObjectNode.class);
        String result = JacksonMapper.getString(objectNode, "Result");
        if (!"success".equalsIgnoreCase(result)) {
            transUtils.writeInfo(ctx, "error", "查询系统设备部件类型失败: " + JacksonMapper.getString(objectNode, "ResultMsg"));
            return;
        }
        // 专业
        ArrayNode array = JacksonMapper.getArray(objectNode, "Content");
        for (JsonNode majorNode : array) {
            ObjectNode major = (ObjectNode) majorNode;
            String majorCode = JacksonMapper.getString(major, "code");
            String majorName = JacksonMapper.getString(major, "class");
            transUtils.writeInfo(ctx, "info", "专业迁移开始: " + majorCode + " " + majorName);

            // 系统
            ArrayNode systemArray = JacksonMapper.getArray(major, "content", JsonNodeFactory.instance.arrayNode());
            for (JsonNode systemNode : systemArray) {
                ObjectNode system = (ObjectNode) systemNode;
                String systemCode = JacksonMapper.getString(system, "code");
                String systemName = JacksonMapper.getString(system, "system");
                transUtils.writeInfo(ctx, "info", "系统迁移开始: " + systemCode + " " + systemName);
                transByType(ctx, "system", systemCode);
                transUtils.writeInfo(ctx, "info", "系统迁移完成: " + systemCode + " " + systemName);

                // 设备
                ArrayNode equipArray = JacksonMapper.getArray(system, "content", JsonNodeFactory.instance.arrayNode());
                for (JsonNode eqiupNode : equipArray) {
                    ObjectNode equip = (ObjectNode) eqiupNode;
                    String equipCode = JacksonMapper.getString(equip, "code");
                    String equipName = JacksonMapper.getString(equip, "facility");
                    transUtils.writeInfo(ctx, "info", "设备迁移开始: " + equipCode + " " + equipName);
                    transByType(ctx, "equipment", equipCode);
                    transUtils.writeInfo(ctx, "info", "设备迁移完成: " + equipCode + " " + equipName);

                    // 部件
                    ArrayNode componentArray = JacksonMapper.getArray(system, "components", JsonNodeFactory.instance.arrayNode());
                    for (JsonNode componentNode : componentArray) {
                        ObjectNode component = (ObjectNode) componentNode;
                        String componentCode = JacksonMapper.getString(equip, "code");
                        String componentName = JacksonMapper.getString(equip, "component");

                        transUtils.writeInfo(ctx, "info", "部件迁移开始: " + componentCode + " " + componentName);
                        transByType(ctx, "component", componentCode);
                        transUtils.writeInfo(ctx, "info", "部件迁移完成: " + componentCode + " " + componentName);
                    }
                }
            }
            transUtils.writeInfo(ctx, "info", "专业迁移完成: " + majorCode + " " + majorName);
        }
    }

    // 迁移关系
    public void transRelation(TransObjectContext ctx) {
        transUtils.transRelation(ctx);
    }

    // 处理失效数据
    public void transInvalidObject(TransObjectContext ctx) {
        transUtils.transInvalidObject(ctx);
    }

    public void error(TransObjectContext context, String message) {
        transUtils.writeInfo(context, "error", message);
    }

    public void over(TransObjectContext context) {
        transUtils.writeInfo(context, "over", null);
        String key = KEY_TRANS + context.getProjectId();
        redis.delete(key);
    }
}
